'''
data = {
    'user_id': [1, 1, 2, 2, 3, 3, 4, 4],
    'item_id': [1, 2, 1, 3, 2, 3, 1, 3],
    'rating': [5, 3, 4, 3, 2, 5, 4, 5]
}

# Create a DataFrame
import pandas as pd
df = pd.DataFrame(data)

# Define the rating scale (e.g., 1 to 5)
reader = Reader(rating_scale=(1, 5))

# Load the data into Surprise's Dataset format
dataset = Dataset.load_from_df(df[['user_id', 'item_id', 'rating']], reader)

# Use KNNBasic for collaborative filtering (user-based)
sim_options = {
    'name': 'cosine',  # Similarity metric
    'user_based': True  # User-based collaborative filtering
}
algo = KNNBasic(sim_options=sim_options)

# Perform cross-validation
cv_results = cross_validate(algo, dataset, measures=['RMSE', 'MAE'], cv=5, verbose=True)

# Print average RMSE and MAE across all folds
print(f"Average RMSE: {cv_results['test_rmse'].mean()}")
print(f"Average MAE: {cv_results['test_mae'].mean()}")

# Train the model on the full dataset
trainset = dataset.build_full_trainset()
algo.fit(trainset)

# Function to recommend items for a specific user
def recommend_items_for_user(algo, user_id, n_recommendations=2):
    # Get a list of all item IDs
    all_items = df['item_id'].unique()
    
    # Get items already rated by the user
    rated_items = df[df['user_id'] == user_id]['item_id'].tolist()
    
    # Predict ratings for unrated items
    recommendations = []
    for item_id in all_items:
        if item_id not in rated_items:
            predicted_rating = algo.predict(user_id, item_id).est
            recommendations.append((item_id, predicted_rating))
    
    # Sort recommendations by predicted rating
    recommendations.sort(key=lambda x: x[1], reverse=True)
    
    # Return top N recommendations
    return recommendations[:n_recommendations]

# Recommend items for user_id = 1
user_id = 1
recommendations = recommend_items_for_user(algo, user_id)
print(f"Recommended items for user {user_id}: {recommendations}")

# Step 1: Represent genres using TF-IDF
tfidf = TfidfVectorizer(stop_words='english')
tfidf_matrix = tfidf.fit_transform(df['genres'])

cosine_sim_df = pd.DataFrame( cosine_similarity(tfidf_matrix, tfidf_matrix), index=df['title'], columns=df['title'])
cosine_sim_df#

cosine_sim_df["Inception"].sort_values(ascending=False).head(5)
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans, AgglomerativeClustering, DBSCAN
from sklearn.metrics import silhouette_score
from scipy.cluster.hierarchy import linkage, dendrogram, fcluster
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import classification_report
from surprise import KNNBasic, Dataset, Reader, accuracy
from surprise.model_selection import train_test_split, cross_validate
from mlxtend.frequent_patterns import apriori, association_rules
 
df.fillna(df.median(), inplace=True)

scaler = StandardScaler()
df_scaled = scaler.fit_transform(df)
pca = PCA()
pca_result = pca.fit(df_scaled)

eigenvalues = pca.explained_variance_
eigenvectors = pca.components_
print("Top 5 Eigenvalues:", eigenvalues[:5])
print("Top 5 Eigenvectors:", eigenvectors[:5])

explained_variance_ratio = np.cumsum(pca.explained_variance_ratio_)
n_components = np.argmax(explained_variance_ratio >= 0.90) + 1
print(f"Number of components explaining 90% variance: {n_components}")

# Transform data using selected number of components
pca_transformed = PCA(n_components=n_components).fit_transform(df_scaled)
pca_df = pd.DataFrame(pca_transformed)

# Determine optimal number of clusters using Elbow Method
wcss = []
for i in range(1, 11):
    kmeans = KMeans(n_clusters=i, random_state=42)
    kmeans.fit(pca_df)
    wcss.append(kmeans.inertia_)
    predict = cluster.fit_predict(X)
    sil.append(silhouette_score(X, predict, random_state= 10))

plt.plot(range(1, 11), wcss, marker='o', linestyle='--')
plt.xlabel('Number of Clusters')
plt.ylabel('WCSS')
plt.title('Elbow Method for Optimal K')
plt.show()

# Finding the optimal K based on WCSS (Elbow Point)
optimal_k_wcss = np.diff(wcss, 2).argmin() + 2  # Second derivative approach

# Finding the optimal K based on highest Silhouette Score
optimal_k_silhouette = K_range[np.argmax(silhouette_scores)]

print(f"Optimal number of clusters using Elbow Method: {optimal_k_wcss}")
print(f"Optimal number of clusters using Silhouette Score: {optimal_k_silhouette}")
# Perform KMeans clustering with optimal clusters (e.g., k=4 as inferred from the Elbow Method)



kmeans = KMeans(n_clusters=optimal_k, random_state=42, n_init=10)
kmeans.fit_predict(X_pca_95)
centroids_df = pd.DataFrame( pca.inverse_transform(kmeans.cluster_centers_), columns=data.feature_names)
centroids_df["Cluster"] = range(optimal_k)

# Print centroid analysis
print("\nCentroids of each cluster (original feature space):\n")
centroids_df.set_index("Cluster").head()
centroids_df.set_index("Cluster").T.plot(kind='bar', figsize=(12, 6))


kmeans = KMeans(n_clusters=4, random_state=42)
kmeans_labels = kmeans.fit_predict(pca_df)

# Evaluate clustering using Silhouette Score
silhouette_avg = silhouette_score(pca_df, kmeans_labels)
print(f'Silhouette Score for KMeans Clustering: {silhouette_avg}')

# Hierarchical clustering dendrograms
linkage_methods = ['single', 'complete', 'average', 'ward']
plt.figure(figsize=(15, 10))
for i, method in enumerate(linkage_methods):
    plt.subplot(2, 2, i+1)
    mergings = linkage(pca_df, method=method, metric='euclidean')
    dendrogram(mergings, truncate_mode='lastp', p=100)
    plt.title(f'Dendrogram ({method} linkage)')
plt.show()

# Load Book_ratings dataset for recommendation system
book_ratings = pd.read_csv('Book_ratings.csv')
print("Total No of Ratings:", book_ratings.shape[0])
print("Total No of Users:", book_ratings['user_id'].nunique())
print("Total No of Books:", book_ratings['book_id'].nunique())

# Popularity-based recommendation
top_books = book_ratings.groupby('book_id').size().sort_values(ascending=False).head(5)
print("Top 5 Recommended Books:", top_books)
reader = Reader(rating_scale=(1, 5))
data = Dataset.load_from_df(book_ratings[['user_id', 'book_id', 'ratings']], reader)
sim_options = {'name': 'cosine', 'user_based': True}
model = KNNBasic(sim_options=sim_options)
cv_results = cross_validate(model, data, measures=['RMSE'], cv=5, verbose=True)
print("Average RMSE:", cv_results['test_rmse'].mean())

# Train the model on the full dataset
trainset = dataset.build_full_trainset()
algo.fit(trainset)

reader = Reader(rating_scale=(book_ratings['ratings'].min(), book_ratings['ratings'].max()))
data = Dataset.load_from_df(book_ratings[['user_id', 'book_id', 'ratings']], reader)
trainset, testset = train_test_split(data, test_size=0.2)
sim_options = {'name': 'cosine', 'user_based': True}
model = KNNBasic(sim_options=sim_options)
model.fit(trainset)
predictions = model.test(testset)
rmse = accuracy.rmse(predictions)
print(f'Collaborative Filtering RMSE: {rmse}')

# Association rule mining
online_retail = pd.read_csv('Online Retail.csv')
online_retail = online_retail.dropna()
france_basket = online_retail[online_retail['Country'] == 'France'].groupby(['InvoiceNo', 'Description'])['Quantity'].sum().unstack().fillna(0)
frequent_itemsets = apriori(france_basket, min_support=0.07, use_colnames=True)
rules = association_rules(frequent_itemsets, metric='lift' ,min_threshold=1)
print(rules)


### **1. K-means Clustering Steps**
K-means is an iterative clustering algorithm that partitions data into **k** clusters. Steps:
1. **Initialize**: Randomly select **k** centroids (cluster centers).
2. **Assign**: Assign each data point to the nearest centroid (using a distance measure like Euclidean distance).
3. **Update**: Recalculate the centroids as the mean of all points in the cluster.
4. **Repeat**: Repeat steps 2 and 3 until centroids no longer change significantly (convergence).


### **2. Cluster Evaluation Metrics**
- **WCSS (Within-Cluster Sum of Squares)**: Measures the compactness of clusters. Lower WCSS indicates tighter clusters.
  \[
  \text{WCSS} = \sum_{i=1}^{k} \sum_{x \in C_i} \|x - \mu_i\|^2
  \]
  where \(C_i\) is the \(i\)-th cluster, \(\mu_i\) is its centroid, and \(x\) is a data point.

- **Silhouette Score**: Measures how similar a point is to its own cluster compared to other clusters. Ranges from -1 to 1, where higher values indicate better clustering.
  \[
  \text{Silhouette Score} = \frac{b - a}{\max(a, b)}
  \]
  where \(a\) is the mean intra-cluster distance, and \(b\) is the mean nearest-cluster distance.

- **Cophenetic Correlation**: Used for hierarchical clustering. Measures how well the dendrogram preserves the pairwise distances between data points. Values closer to 1 indicate better preservation.

---

### **3. Linkage Methods**
Linkage methods define how distances between clusters are calculated in hierarchical clustering:
- **Single Linkage**: Distance between two clusters is the shortest distance between any two points in the clusters.
- **Complete Linkage**: Distance is the longest distance between any two points in the clusters.
- **Average Linkage**: Distance is the average distance between all pairs of points in the clusters.
- **Ward’s Method**: Minimizes the variance when merging clusters.

---

### **4. Market Basket Analysis**
A technique used to identify relationships between products purchased together. Key concepts:
- **Support**: Probability of two items being purchased together.
  \[
  \text{Support}(A \rightarrow B) = P(A \cap B)
  \]
- **Confidence**: Probability of buying item B given that item A is purchased.
  \[
  \text{Confidence}(A \rightarrow B) = \frac{P(A \cap B)}{P(A)}
  \]
- **Lift**: Measures the strength of association between items.
  \[
  \text{Lift}(A \rightarrow B) = \frac{P(A \cap B)}{P(A) \cdot P(B)}
  \]
  - Lift > 1: Positive association.
  - Lift = 1: No association.
  - Lift < 1: Negative association.

---

### **5. PCA and Dimensionality Reduction**
- **Principal Component Analysis (PCA)**: Reduces the dimensionality of data while preserving variance.
  - Steps:
    1. Standardize the data.
    2. Compute the covariance matrix.
    3. Perform eigenvalue decomposition.
    4. Select top \(k\) eigenvectors (principal components) based on eigenvalues.
    5. Project data onto the new subspace.

- **Dimensionality Reduction**: Techniques like PCA, t-SNE, and UMAP reduce the number of features while retaining important information.

---

### **6. Distance Measures and Calculations**
Common distance measures for clustering and similarity:
- **Euclidean Distance**:
  \[
  d(x, y) = \sqrt{\sum_{i=1}^{n} (x_i - y_i)^2}
  \]
- **Manhattan Distance**:
  \[
  d(x, y) = \sum_{i=1}^{n} |x_i - y_i|
  \]
- **Cosine Similarity**:
  \[
  \text{Cosine Similarity} = \frac{x \cdot y}{\|x\| \|y\|}
  \]
- **Mahalanobis Distance**:
  \[
  d(x, y) = \sqrt{(x - y)^T S^{-1} (x - y)}
  \]
  where \(S\) is the covariance matrix.
 

### **(a) Steps for the K-means clustering algorithm**  
1. **Choose the number of clusters (k)**: Determine how many clusters you want to form.  
2. **Initialize cluster centroids**: Randomly select \( k \) points from the dataset as initial cluster centroids.  
3. **Assign points to clusters**: Assign each data point to the nearest cluster centroid based on Euclidean distance (or other distance measures).  
4. **Update centroids**: Compute the mean of all points in each cluster and set the centroid to this mean.  
5. **Repeat steps 3 & 4** until convergence: Stop if centroids no longer change or a set number of iterations is reached.  

---

### **(b) How do you define a group of clusters as a good quality cluster?**  
A good quality clustering should satisfy:  
- **Cohesion (Intra-cluster similarity)**: Points within a cluster should be as close as possible.  
- **Separation (Inter-cluster dissimilarity)**: Different clusters should be well-separated.  
- **Silhouette score**: Measures how similar a point is to its cluster vs. other clusters. A score close to 1 indicates well-defined clusters.  
- **Low inertia**: The sum of squared distances of samples to their closest cluster center should be minimized.  

---

### **(c) Dimensionality Reduction Techniques & Principal Components Computation**  
Dimensionality reduction techniques reduce the number of features while preserving information. Examples:  
- **PCA (Principal Component Analysis)**  
- **t-SNE (t-Distributed Stochastic Neighbor Embedding)**  
- **LDA (Linear Discriminant Analysis)**  
- **Autoencoders**  

**Steps for Computing Principal Components (PCA):**  
1. **Standardize the data**: Normalize each feature to have zero mean and unit variance.  
2. **Compute the covariance matrix**: This represents feature relationships.  
3. **Compute eigenvalues and eigenvectors**: Solve for principal components.  
4. **Sort eigenvectors by eigenvalues**: Higher variance components are prioritized.  
5. **Project data onto selected components**: Reduce dimensions by selecting top eigenvectors.  

---

### **(d) Four Linkage Methods**  
1. **Single Linkage**: Distance between clusters is the minimum distance between any two points.  
2. **Complete Linkage**: Distance between clusters is the maximum distance between any two points.  
3. **Average Linkage**: Distance between clusters is the average of all pairwise distances.  
4. **Centroid Linkage**: Distance between clusters is the distance between their centroids.  

---

### **(e) Market Basket Data & Association Rule Terms**  
Market basket analysis finds associations between frequently bought items.  
- **Support**: Probability of an item appearing in transactions.  
  - \( \text{Support(A)} = \frac{\text{Transactions containing A}}{\text{Total transactions}} \)  
- **Confidence**: Probability that B is bought if A is bought.  
  - \( \text{Confidence(A→B)} = \frac{\text{Support(A ∩ B)}}{\text{Support(A)}} \)  
- **Lift**: Measures how much more likely B is bought when A is bought compared to random chance.  
  - \( \text{Lift(A→B)} = \frac{\text{Confidence(A→B)}}{\text{Support(B)}} \)  

Example: If people buying milk also buy bread with 60% probability, Confidence(Milk → Bread) = 0.6.

---

### **(f) Density-Based Clustering (DBSCAN)**
DBSCAN groups points based on density.  
- **Core points**: Have at least "MinPts" neighbors within "Epsilon" distance.  
- **Border points**: Near a core point but don’t have enough neighbors.  
- **Noise points**: Outliers, not in any cluster.  

Key feature: It can detect clusters of arbitrary shape and identify outliers.

---

### **(g) Principal Component Analysis (PCA) & Applications**  
PCA reduces dimensions while retaining variance.  
**Working Steps:**  
1. Standardize the dataset.  
2. Compute the covariance matrix.  
3. Find eigenvalues & eigenvectors.  
4. Select top k components & project data.  
**Applications:**  
- Image compression  
- Feature extraction in ML  
- Noise reduction  
- Data visualization  

---

### **(h) Content-Based Recommendation System**  
It recommends items based on user preferences and item features.  
**How it Works:**  
1. Creates a user profile based on past interactions.  
2. Computes item similarities using methods like cosine similarity.  
3. Recommends items similar to those the user liked.  
**Pros**: No need for user data, personalized.  
**Cons**: Struggles with new users/items (cold start problem).  

---

### **(i) Market Basket Analysis & Key Techniques**  
It finds relationships between products in transaction data.  
**Techniques:**  
- **Frequent Itemset Mining** (Apriori, FP-Growth)  
- **Association Rules**  
- **Metrics (Support, Confidence, Lift)**  

---

### **(j) Three Distance Measures for Clustering**  
1. **Euclidean Distance**: \( d(A, B) = \sqrt{\sum (A_i - B_i)^2} \)  
2. **Manhattan Distance**: \( d(A, B) = \sum |A_i - B_i| \)  
3. **Cosine Similarity**: \( d(A, B) = \frac{A \cdot B}{||A|| ||B||} \)  

---

### **(k) Role of Silhouette Score**  
It measures how well-defined clusters are:  
\( S = \frac{b - a}{\max(a, b)} \)  
where:  
- \( a \) is average intra-cluster distance  
- \( b \) is average nearest-cluster distance  

Higher scores (near 1) indicate well-separated clusters.

---

### **(l) Distance Computation Using Linkage Methods**  
- **Complete Linkage**: \( d(BC, A) = \max(4,6) = 6 \)  
- **Single Linkage**: \( d(BC, A) = \min(4,6) = 4 \)  

---

### **(m) PCA vs. SVD**  
| Feature | PCA | SVD |  
|---------|-----|-----|  
| Purpose | Dimensionality reduction | Matrix factorization |  
| Decomposition | Covariance matrix | Any matrix |  
| Output | Principal components | U, Σ, V matrices |  

---

### **(n) & (s) Clusters from Dendrogram**  
The number of clusters is determined by cutting the dendrogram at the red line and grouping connected points. (A visual representation is required).

---

### **(t) Supervised vs. Unsupervised Learning**  
| Feature | Supervised Learning | Unsupervised Learning |  
|---------|-------------------|-------------------|  
| Label Availability | Labeled data | Unlabeled data |  
| Example | Classification, Regression | Clustering, PCA |  
| Goal | Learn mapping from input to output | Find hidden patterns |  

---

### **(u) Steps for K-Means Clustering**  
(Same as answer in (a))  

---

### **(v) Need for Linkage Methods & Comparison**  
Linkage methods define how to merge clusters in hierarchical clustering.  
| Linkage | Distance Criterion | Pros | Cons |  
|---------|------------------|------|------|  
| Single | Min distance | Preserves local structure | Sensitive to noise |  
| Complete | Max distance | Well-separated clusters | Can break large clusters |  
| Average | Avg distance | Balanced approach | Computationally heavy |  
| Centroid | Mean of clusters | Intuitive | Can merge distant clusters |  

---

### **(w) Cluster Inertia Computation**  
Inertia = sum of squared distances from points to centroid.  

---

### **(x) Optimal Clusters from Dendrogram**  
Number of clusters = number of groups formed at the red line level.

---
 
'''